/*..........................................................................*/
/*                                                                          */
/*      L a s t W a v e    P a c k a g e 'signal' 2.1                       */
/*                                                                          */
/*      Copyright (C) 1998-2002 Emmanuel Bacry.                             */
/*      email : lastwave@cmap.polytechnique.fr                              */
/*                                                                          */
/*..........................................................................*/
/*                                                                          */
/*      This program is a free software, you can redistribute it and/or     */
/*      modify it under the terms of the GNU General Public License as      */
/*      published by the Free Software Foundation; either version 2 of the  */
/*      License, or (at your option) any later version                      */
/*                                                                          */
/*      This program is distributed in the hope that it will be useful,     */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of      */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       */
/*      GNU General Public License for more details.                        */
/*                                                                          */
/*      You should have received a copy of the GNU General Public License   */
/*      along with this program (in a file named COPYRIGHT);                */
/*      if not, write to the Free Software Foundation, Inc.,                */
/*      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA             */
/*                                                                          */
/*..........................................................................*/



/****************************************************************************/
/*                                                                          */
/*  signal_file.c   Deals with the files                                    */
/*                                                                          */
/****************************************************************************/



#include "lastwave.h"
#include "signals.h"




/*******************************************************
*
* Write a signal in a file  
*
******************************************************/

/*
 * Function to write the y-values of a signal using raw format
 *   signal      : the signal to write
 *   stream      : the file stream
 *   binaryCoding  : either BinaryLittleEndian or BinaryBigEndian or 0 (in that latter case the current cpu mode is used)
 */

void WriteSigRawStream(SIGNAL signal,STREAM s, char binaryCoding,int nsize)
{
  FILE *stream;
  int flagBL;
  void *array;
  int i;
  
  /* Some basic tests on the stream */
  if (s->mode != StreamWrite) Errorf("WriteSigRawStream() : The stream should be an output stream and not an input stream"); 
  if (s->stream == NULL) Errorf("WriteSigRawStream() : You cannot write a signal to standard output"); 
  stream = s->stream; 
  
  /* test on nsize */
  if (nsize != 0 && nsize != sizeof(float) && nsize != sizeof(double)) 
    Errorf("ReadSigRawStream() : Bad number of bytes (%d) for float (must be either %d or %d or 0)",nsize,sizeof(float),sizeof(double)); 
  if (nsize == 0) nsize = sizeof(LWFLOAT);
  
  /* Set the array the data must be read in */
  if (sizeof(LWFLOAT) == nsize) array = signal->Y;
  else {
    array = Malloc(nsize*signal->size);  
    for (i=0;i<signal->size;i++) {
      switch(nsize) {
        case sizeof(float) : ((float *) array)[i] = (float) signal->Y[i]; break;
        case sizeof(double) : ((double *) array)[i] = (double) signal->Y[i]; break;
      }
    }
  }
  
  /* Big/Little conversions */
  if (binaryCoding == 0) {
    if (IsCPULittleEndian) binaryCoding = BinaryLittleEndian;
    else binaryCoding = BinaryBigEndian;
  }
  if (((binaryCoding == BinaryLittleEndian) && IsCPUBigEndian) || ((binaryCoding == BinaryBigEndian) && IsCPULittleEndian)) {
    flagBL = YES;
    BigLittleValues(array,signal->size,nsize);
  }
  else flagBL = NO;
  
  /* We then write the values */	
  fwrite(array,nsize,signal->size,stream);
  
  /* Go back if necessary */
  if (flagBL && sizeof(LWFLOAT) == nsize) {
    BigLittleValues(array,signal->size,nsize);
  }
  
  /* Free the array */
  if (sizeof(LWFLOAT) != nsize) Free(array);
}

/* 
* Same as above but with a filename instead of a stream 
 */

void WriteSigRawFile(SIGNAL signal,char *filename, char binaryCoding,int nsize)
{
  STREAM stream;
  
  /* Open file */
  stream = OpenFileStream(filename,"w");
  
  if (stream == NULL) Errorf("WriteSigRawFile() : Error while opening the file %s",filename);
	
  WriteSigRawStream(signal,stream,binaryCoding,nsize);
  
  CloseStream(stream);
}

/*
 * Function to write a signal using only ascii format
 *   signal      : the signal to write
 *   stream      : the file stream
 *   flagBinary  : If YES then the values are coded using binary codes (and not ascii)
 *   mode == "xy"    ==> write X Y (two columns in ascii mode and X-values followed by Y-values if binary mode)
 *		     "yx"    ==> write Y X (two columns in ascii mode and Y-values followed by X-values if binary mode)
 *			 "x"     ==> write X   (just X is stored)
 *			 "y"     ==> write Y   (just Y is stored)    
 *           ""      ==> the  function chooses the best mode.
 *   flagHeader  : Do we want a header to be written ? (if flagBinary==YES then one should have flagHeader==YES)
 *   format1,format2 : ascii printf type of format (start with %)
 */

void WriteSigStream(SIGNAL signal,STREAM s, char flagBinary, char *mode,int flagHeader,char *format1,char *format2)
{
  int i;
  FILE *stream;
  char *type1,*type2;
  char *format;
  char theFormat[1000];
  
  
  /* Some basic tests on the stream */
  if (s->mode != StreamWrite) Errorf("WriteSigStream() : The stream should be an output stream and not an input stream"); 
  if (s->stream == NULL) Errorf("WriteSigStream() : You cannot write a signal to standard output"); 
  stream = s->stream; 
  
  if ((*format1 != '\0' || *format2 != '\0') && flagBinary) Errorf("WriteSigStream() : Cannot specify a format in binary mode");
  
  /* Test the format1 */
  if (*format1 == '\0') type1 = NULL;
  else {
    format = format1;
    if (strlen(format)<2 && format[0] != '%') Errorf("WriteSigStream() : Bad format '%s'",format1);    
    format++;
    while(*format != '\0' && *format != '%' && !isalpha(*format)) format++;
    if (*format == '\0' || *format == '%') Errorf("WriteSigStream() : Bad format '%s'",format1);
    /* Case of an int */
    if (*format == 'd' || *format == 'i' || *format == 'o' ||
        *format == 'x' || *format == 'X' || *format == 'u') type1 = intType;
    /* Case of an double */
    else if (*format == 'f' || *format == 'e' || *format == 'E' ||
             *format == 'g' || *format == 'G') type1 = floatType;
    else Errorf("WriteSigStream() : Bad format '%s", format1);             
  }
  
  /* Test the X-format */
  if (*format2 == '\0') type2 = NULL;
  else {
    format = format2;
    if (strlen(format)<2 && format[0] != '%') Errorf("WriteSigStream() : Bad format '%s'",format2);    
    format++;
    while(*format != '\0' && *format != '%' && !isalpha(*format)) format++;
    if (*format == '\0' || *format == '%') Errorf("WriteSigStream() : Bad format '%s'",format2);
    /* Case of an int */
    if (*format == 'd' || *format == 'i' || *format == 'o' ||
        *format == 'x' || *format == 'X' || *format == 'u') type2 = intType;
    /* Case of an double */
    else if (*format == 'f' || *format == 'e' || *format == 'E' ||
             *format == 'g' || *format == 'G') type2 = floatType;
    else Errorf("WriteSigStream() : Bad format '%s", format2);             
  }
  
  /* Test the mode */
  if (!strcmp(mode,"")) {
    if (signal->type == YSIG) mode = "y";
    else mode = "xy";
  }
  if (strcmp(mode,"xy") && strcmp(mode,"yx") && strcmp(mode,"x") &&strcmp(mode,"y"))
    Errorf("WriteSigStream() : Mode '%s' not allowed",mode);
  
  /* Some basic checkings */
  if (flagBinary==YES) {
    if (flagHeader ==NO) Errorf("WriteSigStream() : Cannot write binary values without header");
  }
	
  /* Write header if asked */
  if (flagHeader == YES) {
    fprintf(stream,"LastWave Header\n");
    fprintf(stream,"size %d\n",signal->size);
    if (signal->name == NULL || signal->name[0] == '\0') fprintf(stream,"name none\n");
    else fprintf(stream,"name %s\n",signal->name);
    fprintf(stream,"firstp %d\n",signal->firstp);
    fprintf(stream,"lastp %d\n",signal->lastp);
    
    if (flagBinary == NO) fprintf(stream,"binary no\n");
    else {
      if (IsCPULittleEndian) fprintf(stream,"binary little %ld\n",sizeof(LWFLOAT));
      else fprintf(stream,"binary big %ld\n",sizeof(LWFLOAT));
    }
    
    /* 2 types of signal XYSIG or YSIG */
    if (signal->type == YSIG && !strcmp(mode,"y")) {
#ifdef NUMDOUBLE	
      fprintf(stream,"dx %.16g\n",signal->dx);
      fprintf(stream,"x0 %.16g\n",signal->x0);
#else
      fprintf(stream,"dx %.8g\n",signal->dx);
      fprintf(stream,"x0 %.8g\n",signal->x0);
#endif
    }
    if (signal->type == XYSIG && strcmp(mode,"xy") && strcmp(mode,"yx")) {
#ifdef NUMDOUBLE	
      fprintf(stream,"dx %.16g\n",1.0);
      fprintf(stream,"x0 %.16g\n",0.0);
#else
      fprintf(stream,"dx %.8g\n",1.0);
      fprintf(stream,"x0 %.8g\n",0.0);
#endif
    }
    
    fprintf(stream,"End of Header\n");
  }
  
  if (flagBinary == NO) {
    
    if (type1 == NULL) {
      type1 = floatType;
#ifdef NUMDOUBLE	
      format1 = "%.16g";
#else
      format1 = "%.8g";
#endif	
    }
    
    if (type2 == NULL) {
      type2 = floatType;
#ifdef NUMDOUBLE	
      format2 = "%.16g";
#else
      format2 = "%.8g";
#endif	
    }
    
    /* We then write in the file */
    /* according to the mode     */
    if (!strcmp(mode,"y")) {
      sprintf(theFormat,"%s\n",format1);    
      if (type1 == floatType) {  
        for (i=0;i<signal->size;i++) {
          fprintf(stream,theFormat,signal->Y[i]);
        }
      }
      else {  
        for (i=0;i<signal->size;i++) {
          fprintf(stream,theFormat,(int) signal->Y[i]);
        }
      }
    }
    
    if (!strcmp(mode,"x")) {
      sprintf(theFormat,"%s\n",format1);
      if (type1 == floatType) {  
        for (i=0;i<signal->size;i++) {
          fprintf(stream,theFormat,XSig(signal,i));
        }
      } 
      else {
        for (i=0;i<signal->size;i++) {
          fprintf(stream,theFormat,XSig(signal,i));
        }
      }
    }
    
    if (!strcmp(mode,"xy")) {
      sprintf(theFormat,"%s    %s\n",format1,format2);
      if (type1 == floatType) {
        if (type2 == floatType) {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,XSig(signal,i),signal->Y[i]);
          }
        }
        else {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,XSig(signal,i),(int) signal->Y[i]);
          }
        }
      }
      else {
        if (type2 == floatType) {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,(int) XSig(signal,i),signal->Y[i]);
          }
        }
        else {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,(int) XSig(signal,i),(int) signal->Y[i]);
          }
        }
      }
    }	
  	
    if (!strcmp(mode,"yx")) {
      sprintf(theFormat,"%s    %s\n",format1,format2);
      if (type1 == floatType) {
        if (type2 == floatType) {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,signal->Y[i],XSig(signal,i));
          }
        }
        else {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,signal->Y[i],(int) XSig(signal,i));
          }
        }
      }
      else {
        if (type2 == floatType) {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,(int) signal->Y[i],XSig(signal,i));
          }
        }
        else {
          for (i=0;i<signal->size;i++) {
            fprintf(stream,theFormat,(int) signal->Y[i],(int) XSig(signal,i));
          }
        }
      }
    }
  }
  
  else {
    /* We then write the values */	
    if (!strcmp(mode,"xy")) { 
      fwrite(signal->X,sizeof(LWFLOAT),signal->size,stream);
      fwrite(signal->Y,sizeof(LWFLOAT),signal->size,stream);
    }
    else if (!strcmp(mode,"yx")) { 
      fwrite(signal->Y,sizeof(LWFLOAT),signal->size,stream);
      fwrite(signal->X,sizeof(LWFLOAT),signal->size,stream);
    }
    if (!strcmp(mode,"x")) { 
      fwrite(signal->X,sizeof(LWFLOAT),signal->size,stream);
    }
    if (!strcmp(mode,"y")) { 
      fwrite(signal->Y,sizeof(LWFLOAT),signal->size,stream);
    }
  }
}


/* 
* Same as above but with a filename instead of a stream 
 */

void WriteSigFile(SIGNAL signal,char *filename, char flagBinary, char *mode,int flagHeader, char *formatX, char *formatY)
{
  STREAM stream;
  
  /* Open file */
  stream = OpenFileStream(filename,"w");
  
  if (stream == NULL) Errorf("WriteSigFile() : Error while opening the file %s",filename);
	
  WriteSigStream(signal,stream,flagBinary,mode,flagHeader,formatX,formatY);
  
  CloseStream(stream);
}


/*
 * The corresponding command 
 */

void C_Write(char **argv)
{
  SIGNAL signal;
  STREAM stream;
  char *filename,*mode,*bmode,*format1,*format2;
  int flagHeader,flagBinary,flagRaw,nsize;
  char opt,binaryCoding;
  
  argv = ParseArgv(argv,tSIGNALI,&signal,-1);
  argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
  
  if (stream == NULL) argv = ParseArgv(argv,tSTR,&filename,-1);
  else filename = NULL;
  
  
  argv = ParseArgv(argv,tSTR_,"",&mode,-1);
  
  /* options */
  flagHeader = YES;
  flagBinary = YES;
  flagRaw = NO;
  format1 = format2 = "";
  while(opt = ParseOption(&argv)) { 
    switch(opt) {
      case 'h': flagHeader = NO; break;
      case 'f': 
        argv = ParseArgv(argv,tSTR,&format1,-1);
        format2 = format1;
        if (*argv != NULL && **argv != '-') argv = ParseArgv(argv,tSTR,&format2,-1);
          break;
      case 'a': flagBinary = NO; break;
      case 'r':
        if (*argv != NULL && '0'<=**argv && **argv<='9') {
          Printf("hein\n");
          argv = ParseArgv(argv,tINT_,0,&nsize,-1);
          bmode = "";
        }
        else argv = ParseArgv(argv,tSTR_,"",&bmode,tINT_,0,&nsize,-1);
        if (!strcmp(mode,"little")) binaryCoding = BinaryLittleEndian;
        else if (!strcmp(mode,"big")) binaryCoding = BinaryBigEndian;
        else if (!strcmp(mode,"")) binaryCoding = 0;
        else Errorf("Bad binary mode '%s'",mode);
        flagRaw = YES;
        if (nsize != 0 && nsize != sizeof(float) && nsize != sizeof(double)) 
          Errorf("Bad number of bytes (%d) for float (must be either %d or %d)",nsize,sizeof(float),sizeof(double)); 
          break;
      default: ErrorOption(opt);
    }
  }    
  NoMoreArgs(argv);
  
  if (flagRaw == YES) {
    if (*format1 != '\0') Errorf("Cannot specify a format in raw mode");
    if (stream == NULL) WriteSigRawFile(signal,filename,binaryCoding,nsize);
    else WriteSigRawStream(signal,stream,binaryCoding,nsize);
  }
  else {
    if (*format1 != '\0' && flagBinary) Errorf("Cannot specify a format in binary mode");
    if (stream == NULL) WriteSigFile(signal,filename,flagBinary,mode,flagHeader,format1,format2);
    else WriteSigStream(signal,stream,flagBinary,mode,flagHeader,format1,format2);
  }
}



/*******************************************************
*
* Read a signal in a file  
*
******************************************************/


/*
 * Function to read from a stream a signal coded in raw format, i.e., no header and binary (LWFLOAT) values
 *
 * signal      : the signal to read
 * stream      : the file stream
 *
 * firstIndex  : first index in the file to be read (starting from 0)
 * sizeToRead  : the number of points to be read (if -1 everything is read)
 *
 * binaryCoding  : either BinaryLittleEndian or BinaryBigEndian or 0 (in that latter case the current cpu mode is used)
 * nsize        : in case of binary coding, corresponds to number of bytes to use foreach float (should be either sizeof(float), sizeof(double) or 0)
 */

void ReadSigRawStream(SIGNAL signal,STREAM s, int firstIndex,int sizeToRead, char binaryCoding,int nsize)
{
  long int pos,pos1;
  int size;  
  FILE *stream;
  void *array;
  int i;
  
  /* 
    * Some basic tests on the stream 
   */
  if (s->mode != StreamRead) Errorf("ReadSigRawStream() : The stream should be an input stream and not an output stream"); 
  if (s->stream == NULL) Errorf("ReadSigRawStream() : You cannot read a signal to standard input"); 
  stream = s->stream;   
  
  /* test on nsize */
  if (nsize != 0 && nsize != sizeof(float) && nsize != sizeof(double)) 
    Errorf("ReadSigRawStream() : Bad number of bytes (%d) for float (must be either %d or %d or 0)",nsize,sizeof(float),sizeof(double)); 
  if (nsize == 0) nsize = sizeof(LWFLOAT);
  
  /* Compute the size */
  pos = ftell(stream);
  fseek(stream,0,SEEK_END);
  pos1 = ftell(stream);
  fseek(stream,pos-pos1,SEEK_CUR);
  size = (pos1-pos+1)/nsize;
  
  /* Some checkings */
  if (firstIndex >= size) Errorf("ReadSigRawStream() : 'firstIndex' (%d) is too large",firstIndex);
  if (sizeToRead == -1) sizeToRead = size-firstIndex;
  if (sizeToRead+firstIndex > size) Errorf("ReadSigRawStream() : 'sizeToRead' (%d) is too large",sizeToRead);
  
  /* Allocation */
  SizeSignal(signal,sizeToRead,YSIG);
  
  /* Position the stream */
  if (fseek(stream,firstIndex*nsize,SEEK_CUR) != 0)     
    Errorf("ReadSigRawStream() : firstIndex seems to be too large for this file"); 
  
  /* Set the array the data must be read in */
  if (sizeof(LWFLOAT) == nsize) array = signal->Y;
  else array = Malloc(nsize*sizeToRead);  
  
  /* Read the data */
  if (fread(array,nsize,sizeToRead,stream) != sizeToRead)
    Errorf("ReadSigRawStream() : sizeToRead seems to be too large for this file");
  
  /* Big/Little conversions */
  if (binaryCoding == 0) {
    if (IsCPULittleEndian) binaryCoding = BinaryLittleEndian;
    else binaryCoding = BinaryBigEndian;
  }
  if (((binaryCoding == BinaryLittleEndian) && IsCPUBigEndian) || ((binaryCoding == BinaryBigEndian) && IsCPULittleEndian)) {
    BigLittleValues(array,sizeToRead,nsize);
  }
  
  /* Conversion of the array if necessary */
  if (sizeof(LWFLOAT) != nsize) {
    for (i=0;i<sizeToRead;i++) {
      if (sizeof(float) == nsize) signal->Y[i] = ((float *) array)[i];
      else signal->Y[i] = ((double *) array)[i];
    }
    Free(array);
  }
  
  
  /* Some settings */              
  SetNameSignal(signal,NULL);
}


/* Same as above but with a filename instead of a stream */
void ReadSigRawFile(SIGNAL signal,char *filename,  int firstIndex,int sizeToRead, char binaryCoding, int nsize)
{
  STREAM stream;
  
  /* Open file */
  stream = OpenFileStream(filename,"r");
  if (stream == NULL) Errorf("ReadSigRawFile() : Error while opening the file %s",filename);
  
  ReadSigRawStream(signal,stream,firstIndex,sizeToRead,binaryCoding,nsize);
	
  CloseStream(stream);	
  
  if (!strcmp(signal->name,"")) SetNameSignal(signal,filename);
}

/*
 * Function to get information about a signal stored in a stream without reading the signal itself.
 *
 * The parameters are :
 *
 * stream      : the file stream
 *
 * The following parameters will be filled by te function
 *
 * siginfo : a signal structure where all the fields are filled up except the Y or X
 *           (contains, size, firstp, lastp, XYSIG, YSIG, dx, x0 info)
 *           WARNING : NO ALLOCATION IN 'siginfo' IS MADE
 * header : YES if there is a header
 * flagBinary : YES or NO
 * binaryCoding : BinaryLittleEndian or BinaryBigEndian (if flagBinary == YES only)
 * nCols   : the number of columns (if flagBinary == NO only)
 * nSize : in case of binary coding the number of bytes used to code a float
 *
 * It returns YES if the format is a LastWave format otherwise it returns NO
 *
 * This function does not change the position of the stream.
 */

char ReadInfoSigStream(STREAM s, SIGNAL siginfo, char *header, char *flagBinary, char *binaryCoding, int *nColumns, int *nsize)
{
  int flagHeader;
  int flagXY;
  int firstp,lastp,size;
  LWFLOAT dx,x0,f,f1;
  char name[255],str[1000],*str1,c;
  int n,i;
  int nCol;
  FILE *stream;
  int beg;
  char flagBinary1;
  char flagLittleEndian;
  char flagNewFormat;
  
  /* 
    * Some basic tests on the stream 
   */
  if (s->mode != StreamRead) Errorf("ReadInfoSigStream() :The stream should be an input stream and not an output stream"); 
  if (s->stream == NULL) Errorf("ReadInfoSigStream() : You cannot read a signal to standard input"); 
  stream = s->stream;   
  
  /* 
    * Let's try to read the header 
   */
  flagHeader = NO;
  flagBinary1 = NO;
  flagXY = YES;
  flagNewFormat = NO;
  beg = ftell(stream);
  str1 = "LastWave Header";
  while (*str1 != '\0') {
    c = fgetc(stream);
    if (c != *str1) break;
    str1++;
  }
  
  if (*str1 == '\0') {
    c = fgetc(stream);
    c = fgetc(stream);
    if (c != '\n') ungetc(c,stream);
    flagHeader = YES;
    n = 0;
    n += fscanf(stream,"size %d ",&size);
    n += fscanf(stream,"name %[^\n\r] ",name);
    n += fscanf(stream,"firstp %d ",&firstp);
    n += fscanf(stream,"lastp %d ",&lastp);
		
    if (n != 4) Errorf("ReadSigStream() : Error in the header of the file");
    
    /* Extended header (Signal Package, version 1.5) */
    n += fscanf(stream,"binary %[^\n\r] ",&str);
    if (n == 5) {
      n--;
      flagNewFormat = YES;
      if (!strcmp(str, "no")) flagBinary1 = NO;
      else {
        flagBinary1 = YES;
        if (!strncmp(str, "little",6)) {
          flagLittleEndian = YES; 
          str1 = str+6;
        }
        else if (!strncmp(str, "big",3)) {
          flagLittleEndian = NO;
          str1 = str+3;
        } 
        else Errorf("ReadSigInfoStream() : Bad field value '%s' of field 'binary'",str);
        *nsize = sizeof(float);
        if (*str1 == ' ') {
          str1++;
          sscanf(str1,"%d",nsize);
        }
      } 
    }
    else flagBinary1 = NO;
    /* End of extended header */
    
    fscanf(stream,"%[^\n\r] ",str);
    if (strcmp(str,"End of Header")) { /* it is a YSIG */
               flagXY = NO;
#ifdef NUMDOUBLE
               n += sscanf(str,"dx %lf",&dx);
               n += fscanf(stream,"x0 %lf ",&x0);
#else
               n += sscanf(str,"dx %f",&dx);
               n += fscanf(stream,"x0 %f ",&x0);
#endif
               if (n != 6) Errorf("ReadSigStream() : Error in the header of the file");
               fscanf(stream,"%[^\n\r] ",str);
               if (strcmp(str,"End of Header")) Errorf("ReadSigStream() : Error in the header of the file");
    }
  }

  fseek(stream,beg,SEEK_SET);

/*
 * Case there is a header 
 */
if (flagHeader) {
  *header = YES;
  siginfo->size = size;
  siginfo->lastp = lastp;
  siginfo->firstp = firstp;
  if (flagXY == NO) {
    siginfo->dx = dx;
	  siginfo->x0 = x0;
	  siginfo->type = YSIG;
	  *nColumns = 1;
  }
  else {
    siginfo->type = XYSIG;
    *nColumns = 2;
  }
  
  if (flagBinary1) {
    *flagBinary = YES;
    if (flagLittleEndian) *binaryCoding = BinaryLittleEndian;
    else  *binaryCoding = BinaryBigEndian;
  }
  
  return(YES);
}


/*
 * Case there is no header
 */

/* How many columns on the first line ? */
#ifdef NUMDOUBLE
if (fscanf(stream,"%lf",&f) != 1) return(NO);
#else
if (fscanf(stream,"%f",&f) != 1) return(NO);
#endif
  fseek(stream,beg,SEEK_SET);  
fscanf(stream,"%[^\n\r] ",str);
nCol = 0;
str1 = str;
while (*str1 != '\0' && (*str1 == ' ' || *str1 == '\t')) str1++;
while (*str1 != '\0') {
  nCol++;
  while (*str1 != '\0' && *str1 != ' ' && *str1 != '\t') str1++;
  while (*str1 != '\0' && (*str1 == ' ' || *str1 == '\t')) str1++;
}
  fseek(stream,beg,SEEK_SET);

if (nCol <= 0) return(NO);

/* Let's read !! */
size = 0;
for(i= 0 ;;i++) { 
  
  if (nCol == 1) {
#ifdef NUMDOUBLE
    n = fscanf(stream,"%lf ",&f);
#else
    n = fscanf(stream,"%f ",&f);
#endif
    if (n == EOF) break;
    if (n != 1) return(NO);
    size++;
  } 
  else if (nCol == 2) {
#ifdef NUMDOUBLE
    n = fscanf(stream,"%lf %lf ",&f,&f1);
#else
    n = fscanf(stream,"%f %f ",&f,&f1);
#endif
    if (n == EOF) break;
    if (n != 2) return(NO);
    size++;
  } 
  else {
#ifdef NUMDOUBLE
    n = fscanf(stream,"%lf %lf%*[^\n\r] ",&f,&f1);
#else
    n = fscanf(stream,"%f %f%*[^\n\r] ",&f,&f1);
#endif
    if (n == EOF) break;
    if (n != 2) return(NO);
    size++;
  } 
}

  fseek(stream,beg,SEEK_SET);
/* Fill up the results */
*header = NO;
siginfo->size = size;
siginfo->lastp = size-1;
siginfo->firstp = 0;
siginfo->dx = 1;
siginfo->x0 = 0;
if (nCol == 2) siginfo->type = XYSIG;
else siginfo->type = YSIG;
*nColumns = nCol;
*flagBinary = NO;

return(YES);
}

/* Same as above but with a filename instead of a stream */
char ReadInfoSigFile(char *filename, SIGNAL siginfo, char *header, char *flagBinary, char *binaryCoding, int *nColumns, int *nsize)
{
  STREAM stream;
  char ans;
  
  /* Open file */
  stream = OpenFileStream(filename,"r");
  if (stream == NULL) Errorf("ReadInfoSigFile() : Error while opening the file %s",filename);
  
  ans = ReadInfoSigStream(stream,siginfo,header,flagBinary,binaryCoding,nColumns,nsize);
	
  CloseStream(stream);	  
  
  return(ans);
}

/*
 * Function to read from a stream  a signal.
 * It can be coded using 
 *     - (Optional LastWave Header), ascii coded floats
 *     - Header, binary coded floats
 *
 * The parameters are :
 *
 * signal      : the signal to read
 * stream      : the file stream
 *
 * firstIndex  : first index in the file to be read (starting from 0)
 * sizeToRead  : the number of points to be read (if -1 everything is read)
 *
 * The two next fields are used only for ascii streams (othewise they should be set to 0,0)
 *
 * xcol        : The column number corresponding to the x-values (starting from 1)
 * ycol        : The column number corresponding to the y-values (starting from 1)
 *    If xcol or ycol is 0 then the function make some natural assumptions.....
 *    If xcol == -1 then only the Y field should be read
 */

void ReadSigStream(SIGNAL signal,STREAM s, int firstIndex,int sizeToRead, int xcol,int ycol)
{
  int flagHeader;
  int flagXY;
  int firstp,lastp,size;
  LWFLOAT dx,x0,f;
  char name[255],str[1000],*str1,c;
  int n,i,j;
  int nCol;
  FILE *stream;
  fpos_t beginning;
  long int beg;
  char flagBinary;
  char flagLittleEndian;
  char flagNewFormat;
  void *array;
  size_t nsize;
  
  
  /* 
    * Some basic tests on the stream 
   */
  if (s->mode != StreamRead) Errorf("ReadSigStream() :The stream should be an input stream and not an output stream"); 
  if (s->stream == NULL) Errorf("ReadSigStream() : You cannot read a signal to standard input"); 
  stream = s->stream;   
  
  /* 
    * Let's try to read the header 
   */
  flagHeader = NO;
  flagBinary = NO;
  flagXY = YES;
  flagNewFormat = NO;
  beg = ftell(stream);
  str1 = "LastWave Header";
  while (*str1 != '\0') {
    c = fgetc(stream);
    if (c != *str1) break;
    str1++;
  }
  
  if (*str1 == '\0') {
    c = fgetc(stream);
    c = fgetc(stream);
    if (c != '\n') ungetc(c,stream);
    flagHeader = YES;
    n = 0;
    n += fscanf(stream,"size %d ",&size);
    n += fscanf(stream,"name %[^\n\r] ",name);
    n += fscanf(stream,"firstp %d ",&firstp);
    n += fscanf(stream,"lastp %d ",&lastp);
    if (n != 4) Errorf("ReadSigStream() : Error in the header of the file");
    
    n += fscanf(stream,"binary %[^\n\r] ",&str);
    if (n == 5) {
      n--;
      flagNewFormat = YES;
      if (!strcmp(str, "no")) flagBinary = NO;
      else {
        flagBinary = YES;
        if (!strncmp(str, "little",6)) {
          flagLittleEndian = YES; 
          str1 = str+6;
        }
        else if (!strncmp(str, "big",3)) {
          flagLittleEndian = NO;
          str1 = str+3;
        } 
        else Errorf("ReadSigStream() : Bad field value '%s' of field 'binary'",str);
        nsize = sizeof(float);
        if (*str1 == ' ') {
          str1++;
          sscanf(str1,"%ld",&nsize);
        }
        if (nsize != sizeof(float) && nsize != sizeof(double)) Errorf("ReadSigStream() : Can't read binary files using %d bytes to store numbers (that size is neither float or double)",nsize);
      } 
    }
    else flagBinary = NO;
    /* End of extended header */
    
    fscanf(stream,"%[^\n\r] ",str);
    if (strcmp(str,"End of Header")) { /* it is a YSIG */
               flagXY = NO;
#ifdef NUMDOUBLE
               n += sscanf(str,"dx %lf ",&dx);
               n += fscanf(stream,"x0 %lf ",&x0);
#else
               n += sscanf(str,"dx %f ",&dx);
               n += fscanf(stream,"x0 %f ",&x0);
#endif
               if (n != 6) Errorf("ReadSigStream() : Error in the header of the file");
      if (!flagBinary) {
               fscanf(stream,"%[^\n\r] ",str);
               if (strcmp(str,"End of Header")) Errorf("ReadSigStream() : Error in the header of the file");
    }
      else {
        str1 = "End of Header\n";
        while (*str1 != '\0') {
          if (fgetc(stream) != *str1) break;
          str1++;
        }
        if (*str1 != '\0') Errorf("ReadSigStream() : Error in the header of the file");
      }
    }
    beg = ftell(stream);
  }
  fseek(stream,beg,SEEK_SET);




/* CASE OF AN ASCII FILE */
if (flagBinary == NO) {
  
  while (1) {
    
    /* 
    * How many columns on the first line ? 
     */
#ifdef NUMDOUBLE
    if (fscanf(stream,"%lf",&f) != 1) Errorf("ReadSigStream() : Bad format");
#else
    if (fscanf(stream,"%f",&f) != 1) Errorf("ReadSigStream() : Bad format");
#endif
      fseek(stream,beg,SEEK_SET);  
    
    fscanf(stream,"%[^\n\r] ",str);
    nCol = 0;
    str1 = str;
    while (*str1 != '\0' && (*str1 == ' ' || *str1 == '\t')) str1++;
    while (*str1 != '\0') {
      nCol++;
      while (*str1 != '\0' && *str1 != ' ' && *str1 != '\t') str1++;
      while (*str1 != '\0' && (*str1 == ' ' || *str1 == '\t')) str1++;
    }
    
    if (nCol == 0) {
      if (flagNewFormat) Errorf("ReadSigStream() : Cannot figure out the number of colums");
      break;
    }
    
    /* 
      * Possible errors and set xcol and ycol
     */  
    if (xcol > nCol) Errorf("ReadSigStream() : Only %d columns in file ==> x column number %d is incorrect",nCol,xcol);
    if (ycol > nCol) Errorf("ReadSigStream() : Only %d columns in file ==> y column number %d is incorrect",nCol,ycol);
    if (xcol > 0 && ycol <= 0) Errorf("ReadSigStream() : Sorry you cannot read just X");
    if (xcol <= 0) {
      if (xcol == -1) xcol = 0;
      else if (flagHeader) {
        if (flagXY) xcol = 1;
        else xcol = 0;
      }
      else if (nCol == 2) xcol = 1;
      else xcol = 0;
    }
    if (ycol <= 0) {
      if (flagHeader) {
        if (flagXY) ycol = 2;
        else ycol = 1;
      }
      else if (nCol == 2) ycol = 2;
      else ycol = 1;
    }
    
    /* 
      * Compute the size of the signal if no header and size to read is not specified
     */
    if (flagHeader == NO && sizeToRead == -1) {
        fseek(stream,beg,SEEK_SET);
      size = 0;
      while(fscanf(stream,"%[^\n\r] ",str) != EOF) size++;
    }
    
    /* 
      * Check firstIndex and sizeToRead 
     */
    if (firstIndex < 0) firstIndex = 0;
    if (sizeToRead == -1) size = size - firstIndex;
    else if (flagHeader == YES) {
      if (firstIndex+sizeToRead > size) Errorf("ReadSigStream() : Either firstIndex or sizeToRead is too big");
      size = sizeToRead;
    }
    else size = sizeToRead;
    
    
    /* 
      * Allocation 
     */		
    if (xcol != 0) SizeSignal(signal,size,XYSIG);
    else SizeSignal(signal,size,YSIG);           	
		
    /* 
      * We skip everything up to firstIndex
     */
      fseek(stream,beg,SEEK_SET);
    for (i = 0;i<firstIndex;i++) {
      n = fscanf(stream,"%[^\n\r] ",str);
      if (n!=1) {
        if (flagNewFormat) Errorf("ReadSigStream() : Signal is too short while reading file");
        break;
      }
    }
    
    /* 
      * Let's read !!
     */
    for(i= 0 ;i<size;i++) {
      
      /* Case both X and Y must be read */
      if (xcol != 0) {
        
        for (j = 1; j<MIN(xcol,ycol);j++) fscanf(stream,"%*s");
        
        n = 0;
#ifdef NUMDOUBLE
        if (xcol < ycol) n = fscanf(stream,"%lf",&(signal->X[i]));
        else if (xcol > ycol) n = fscanf(stream,"%lf",&(signal->Y[i]));
        else {
          n = fscanf(stream,"%lf%*[^\n\r] ",&(signal->Y[i]));
#else
          if (xcol < ycol) n = fscanf(stream,"%f",&(signal->X[i]));
          else if (xcol > ycol) n = fscanf(stream,"%f",&(signal->Y[i]));
          else {
            n = fscanf(stream,"%f%*[^\n\r] ",&(signal->Y[i]));
#endif
            signal->X[i] = signal->Y[i];
            n+=1;
          }
          if (xcol != ycol) {    
            j++;
            for (; j<MAX(xcol,ycol);j++) fscanf(stream,"%*s");
#ifdef NUMDOUBLE
            if (xcol < ycol) n += fscanf(stream,"%lf%*[^\n\r] ",&(signal->Y[i]));
            else n += fscanf(stream,"%lf%*[^\n\r] ",&(signal->X[i]));
#else
            if (xcol < ycol) n += fscanf(stream,"%f%*[^\n\r] ",&(signal->Y[i]));
            else n += fscanf(stream,"%f%*[^\n\r] ",&(signal->X[i]));
#endif
          }  
          if (n != 2) {
            if (flagNewFormat) Errorf("ReadSigStream() : Error while reading file"); 
            break;
          }
        }
        
        /* Case just Y must be read */
        else {
          for (j = 1; j<ycol;j++) fscanf(stream,"%*s");
#ifdef NUMDOUBLE
          n = fscanf(stream,"%lf%*[^\n\r] ",&(signal->Y[i]));
#else
          n = fscanf(stream,"%f%*[^\n\r] ",&(signal->Y[i]));
#endif
          if (n != 1) {
            if (flagNewFormat) Errorf("ReadSigStream() : Error while reading file"); 
            break;
          } 
        } 
      }
      
      if (i != size) break;	
      
      if (flagNewFormat == NO && flagHeader) 
        Warningf("ReadSigStream() : (old Signal Package 1.1.2 header) You should update the format (writing the file again using 'write')");
      
      
      /* Some initialization */
      signal->size = size;
      if (flagHeader) {
        if (sizeToRead == -1){
          signal->lastp = lastp;
          signal->firstp = firstp;
        }
        if (flagXY == NO) {
          signal->dx = dx;
          signal->x0 = x0 + dx*firstIndex;
        }
        if (strcmp(name,"none")) SetNameSignal(signal,name);
        else SetNameSignal(signal,NULL);
      }
      else	{
        signal->x0 = firstIndex;
      }
      
      
      /* We might need to sort the signal */
      if (xcol != 0) SortSig(signal);
      
      return;
    }
  }
  
  /* 
  * If there was an error assuming data is ascii coded  maybe it is because the header is a Signal Package 1.1.2 header
   * and we are currently dealing with binary coded values without knowing it !
   */
  if (flagBinary == NO) {
    if (IsCPULittleEndian) {
      flagLittleEndian = YES;
      Warningf("ReadSigStream() : (old Signal Package 1.1.2 header) You should update the format (writing the file again using 'write'). I am assuming the values are binary coded (little endian).");
    }
    else {
      flagLittleEndian = NO;
      Warningf("ReadSigStream() : (old Signal Package 1.1.2 header) You should update the format (writing the file again using 'write'). I am assuming the values are binary coded (big endian).");
    }
    flagBinary = YES;
      fseek(stream,beg,SEEK_SET);
  }
  
  
  /* CASE OF A BINARY FILE  (there should be a header, otherwise you should use the ReadSigRawStream function) */
  
  if (flagHeader == NO) Errorf("ReadSigStream() : It looks like the stream contains binary coded values with no header. I can't read that");
  
  /* 
  * Check firstIndex and sizeToRead 
   */
  if (firstIndex < 0) firstIndex = 0;
  if (sizeToRead == -1) sizeToRead = size - firstIndex;
  if (firstIndex+sizeToRead > size) Errorf("ReadSigStream() : Either firstIndex or sizeToRead is too big");
  
  /* Case XY */  
  if (flagXY) {
    
    /* Set signal size */
    SizeSignal(signal,sizeToRead,XYSIG);
    if (fseek(stream,firstIndex*nsize,SEEK_CUR) != 0)     
      Errorf("ReadSigStream() : firstIndex seems to be too large for this file");
    
    /* Set the array the data must be read in */
    if (sizeof(LWFLOAT) == nsize) array = signal->X;
    else array = Malloc(nsize*sizeToRead);  
    
    /* Read the data */
    if (fread(array,sizeof(LWFLOAT),sizeToRead,stream) != sizeToRead)
      Errorf("ReadSigStream() : sizeToRead seems to be too large for this file");
    
    /* Big/Little conversions */
    if ((flagLittleEndian && IsCPUBigEndian) || (!flagLittleEndian && IsCPULittleEndian))
      BigLittleValues(array,sizeToRead,nsize);
    
    /* Conversion of the array if necessary */
    if (sizeof(LWFLOAT) != nsize) {
      for (i=0;i<sizeToRead;i++) {
        if (sizeof(float) == nsize) signal->X[i] = ((float *) array)[i];
        else signal->X[i] = ((double *) array)[i];
      }
      Free(array);
    }
    
    /* Move to the Y data */  
    fseek(stream,size-sizeToRead-firstIndex,SEEK_CUR);
    fseek(stream,firstIndex*nsize,SEEK_CUR);
    
    /* Set the array the data must be read in */
    if (sizeof(LWFLOAT) == nsize) array = signal->Y;
    else array = Malloc(nsize*sizeToRead);  
    
    /* Read the data */
    if (fread(array,sizeof(LWFLOAT),sizeToRead,stream) != sizeToRead)
      Errorf("ReadSigStream() : sizeToRead seems to be too large for this file");
    
    /* Big/Little conversions */
    if ((flagLittleEndian && IsCPUBigEndian) || (!flagLittleEndian && IsCPULittleEndian))
      BigLittleValues(array,sizeToRead,nsize);
    
    /* Conversion of the array if necessary */
    if (sizeof(LWFLOAT) != nsize) {
      for (i=0;i<sizeToRead;i++) {
        if (sizeof(float) == nsize) signal->Y[i] = ((float *) array)[i];
        else signal->Y[i] = ((double *) array)[i];
      }
      Free(array);
    }
    
    /* Go to the end of the signal */
    fseek(stream,size-sizeToRead-firstIndex,SEEK_CUR);
  }
  
  /* Case Y */  
  else {
    
    /* Set signal size */
    SizeSignal(signal,sizeToRead,YSIG);
    if (fseek(stream,firstIndex*nsize,SEEK_CUR) != 0)     
      Errorf("ReadSigStream() : firstIndex seems to be too large for this file");
    
    /* Set the array the data must be read in */
    if (sizeof(LWFLOAT) == nsize) array = signal->Y;
    else array = Malloc(nsize*sizeToRead);  
    
    /* Read the data */
    if (fread(array,nsize,sizeToRead,stream) != sizeToRead)
      Errorf("ReadSigStream() : sizeToRead seems to be too large for this file");
    
    /* Big/Little conversions */
    if ((flagLittleEndian && IsCPUBigEndian) || (!flagLittleEndian && IsCPULittleEndian))
      BigLittleValues(array,sizeToRead,nsize);
    
    /* Conversion of the array if necessary */
    if (sizeof(LWFLOAT) != nsize) {
      for (i=0;i<sizeToRead;i++) {
        if (sizeof(float) == nsize) signal->Y[i] = ((float *) array)[i];
        else signal->Y[i] = ((double *) array)[i];
      }
      Free(array);
    }
    
    /* Go to the end of the signal */
    fseek(stream,size-sizeToRead-firstIndex,SEEK_CUR);
  }
  
  /* Some settings */              
  if (strcmp(name,"none")) SetNameSignal(signal,name);
  else SetNameSignal(signal,NULL);
  if (flagXY == NO) {
    signal->dx = dx;
    signal->x0 = x0;
  }
  signal->firstp = firstp;
  signal->lastp = lastp;
}


/* Same as above but with a filename instead of a stream */
void ReadSigFile(SIGNAL signal,char *filename,  int firstIndex,int sizeToRead, int xcol,int ycol)
{
  STREAM stream;
  
  /* Open file */
  stream = OpenFileStream(filename,"r");
  if (stream == NULL) Errorf("ReadSigFile() : Error while opening the file %s",filename);
  
  ReadSigStream(signal,stream,firstIndex,sizeToRead, xcol,ycol);
	
  CloseStream(stream);	
  
  if (!strcmp(signal->name,"")) SetNameSignal(signal,filename);
}


/*
 * The corresponding commands 
 */

void C_Read(char **argv)
{
	SIGNAL signal;
	char *filename,*mode;
  int xcol,ycol,nsize;
	char flagHeader,flagRaw,binaryCoding;
	int firstIndex,sizeToRead;
  char opt;
  STREAM stream;
  
	argv = ParseArgv(argv,tSIGNAL,&signal,-1);
	
  argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
  
  if (stream == NULL) argv = ParseArgv(argv,tSTR,&filename,-1);
  else filename = NULL;
  
  argv = ParseArgv(argv,tINT_,-2,&xcol,tINT_,-2,&ycol,-1);
	
	if (xcol == -2 && ycol == -2) xcol = ycol = 0;
	else if (ycol == -2) {
	  ycol = xcol;
	  xcol = -1;
	}
	else if (xcol < 0 || ycol < 0) Errorf("Bad <xcol> or <ycol> value");
	
	firstIndex = 0;
	sizeToRead = -1;
	
  /* options */
  flagHeader = YES;
  flagRaw = NO;
  while(opt = ParseOption(&argv)) { 
    switch(opt) {
      case 'f':
        argv = ParseArgv(argv,tINT,&firstIndex,-1);
        if (firstIndex < 0) ErrorUsage1(); 
          break;
      case 's': 
        argv = ParseArgv(argv,tINT,&sizeToRead,-1);
        if (sizeToRead <= 0) ErrorUsage1(); 
          break;
      case 'r':
        if (*argv != NULL && '0'<=**argv && **argv<='9') {
          argv = ParseArgv(argv,tINT_,0,&nsize,-1);
          mode = "";
        }
        else argv = ParseArgv(argv,tSTR_,"",&mode,tINT_,0,&nsize,-1);
        if (!strcmp(mode,"little")) binaryCoding = BinaryLittleEndian;
        else if (!strcmp(mode,"big")) binaryCoding = BinaryBigEndian;
        else if (!strcmp(mode,"")) binaryCoding = 0;
        else Errorf("Bad binary mode '%s'",mode);
        flagRaw = YES;
        if (nsize != 0 && nsize != sizeof(float) && nsize != sizeof(double)) 
          Errorf("Bad number of bytes (%d) for float (must be either %d or %d)",nsize,sizeof(float),sizeof(double)); 
          break;
      default: ErrorOption(opt);
    }
  }    
  NoMoreArgs(argv);
  
  if (flagRaw) {
    if (stream == NULL) ReadSigRawFile(signal,filename,firstIndex,sizeToRead,binaryCoding,nsize);
    else ReadSigRawStream(signal,stream,firstIndex,sizeToRead,binaryCoding,nsize);
    return;
  }
  
  if (stream == NULL) ReadSigFile(signal,filename,firstIndex,sizeToRead,xcol,ycol);
  else ReadSigStream(signal,stream,firstIndex,sizeToRead,xcol,ycol);
}

void C_ReadInfo(char **argv)
{
	struct signal siginfo;
	char *filename,header,flagBinary,binaryCoding,*type;
	int nCols,nsize;
	char flagPrint;
  char opt;
  STREAM stream;
  LISTV lv,lv1;
  
  argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
  
  if (stream == NULL) argv = ParseArgv(argv,tSTR,&filename,-1);
  else filename = NULL;
  
  
  /* options */
  flagPrint = NO;
  while(opt = ParseOption(&argv)) { 
    switch(opt) {
      case 'p':
        flagPrint = YES;
        break;
      default: ErrorOption(opt);
    }
  }    
  NoMoreArgs(argv);
  
  if (filename) {
    if (ReadInfoSigFile(filename, &siginfo, &header,&flagBinary, &binaryCoding, &nCols,&nsize)==NO) {
      if (flagPrint) Printf("Filename '%s' cannot be read using the 'read' command\n",filename);
      return;
    }
  }
  else {
    if (ReadInfoSigStream(stream, &siginfo, &header,&flagBinary, &binaryCoding, &nCols,&nsize)==NO) {
      if (flagPrint) Printf("Filename '%s' cannot be read using the 'read' command\n",filename);
      return;
    }
  }  
  
  if (flagPrint) {
    if (filename) {
      if (header) Printf("Filename '%s' has a LastWave header\n",filename);
      else Printf("Filename '%s' has no header\n",filename);
    }
    else {
      if (header) Printf("Stream has a LastWave header\n");
      else Printf("Stream has no header\n");
    }
    if (header) {
      if (siginfo.type == YSIG) {
        if (flagBinary == YES) {
          if (binaryCoding == BinaryLittleEndian) Printf("YSIG (Binary coded, little endian)\n");
          else Printf("YSIG (Binary coded, big endian)\n");
          if (nsize == sizeof(float)) type = "float";
          else if (nsize == sizeof(double)) type = "double";
          else type = "???";
          Printf("Float size is %d bytes (C-type is '%s')\n",nsize,type); 
        }
        else  Printf("YSIG (ascii coded)\n");
        Printf("size   : %d\n",siginfo.size);
#ifdef NUMDOUBLE         
        Printf("x0     : %.16g\n",siginfo.x0);
        Printf("dx     : %.16g\n",siginfo.dx);
#else
        Printf("x0     : %.8g\n",siginfo.x0);
        Printf("dx     : %.8g\n",siginfo.dx);
#endif
        Printf("firstp : %d\n",siginfo.firstp);
        Printf("lastp  : %d\n",siginfo.lastp);
      }
      else {
        if (flagBinary == YES) {
          if (binaryCoding == BinaryLittleEndian) Printf("XYSIG (Binary coded, little endian)\n");
          else Printf("XYSIG (Binary coded, big endian)\n");
          if (nsize == sizeof(float)) type = "float";
          else if (nsize == sizeof(double)) type = "double";
          else type = "???";
          Printf("Float size is %d bytes (C-type is '%s')\n",nsize,type); 
        }
        else  Printf("XYSIG (ascii coded)\n");
        Printf("size   : %d\n",siginfo.size);
        Printf("firstp : %d\n",siginfo.firstp);
        Printf("lastp  : %d\n",siginfo.lastp);
      }
    }
    else {
      Printf("%d Columns (ascii coded)\n",nCols);
      Printf("size   : %d\n",siginfo.size);
    }
  }
  
  else {
    lv = TNewListv();
    SetResultValue(lv);
    if (header) {
      if (siginfo.type == YSIG) {
        AppendInt2Listv(lv,1);
        if (flagBinary == YES) {
          if (binaryCoding==BinaryLittleEndian) AppendStr2Listvf(lv,"binary little %d",nsize);
          else AppendStr2Listvf(lv,"binary big %d",nsize);
        }
        else {
          AppendStr2Listv(lv,"ascii");
        }
        AppendInt2Listv(lv,siginfo.size);
        lv1 = TNewListv();
        AppendValue2Listv(lv,(VALUE) lv1);
        AppendStr2Listv(lv1,"y");
        AppendFloat2Listv(lv1,siginfo.x0);
        AppendFloat2Listv(lv1,siginfo.dx);
        lv1 = TNewListv();
        AppendValue2Listv(lv,(VALUE) lv1);
        AppendFloat2Listv(lv1,siginfo.firstp);
        AppendFloat2Listv(lv1,siginfo.lastp);
      }
      else {
        if (flagBinary == YES) {
          AppendInt2Listv(lv,1);
          if (binaryCoding==BinaryLittleEndian) AppendStr2Listvf(lv,"binary little %d",nsize);
          else AppendStr2Listvf(lv,"binary big %d",nsize);
        }
        else {
          AppendStr2Listv(lv,"ascii");
        }
        AppendInt2Listv(lv,siginfo.size);
        AppendStr2Listv(lv,"xy");
        lv1 = TNewListv();
        AppendValue2Listv(lv,(VALUE) lv1);
        AppendFloat2Listv(lv1,siginfo.firstp);
        AppendFloat2Listv(lv1,siginfo.lastp);
      }
    }
    else {      
      AppendInt2Listv(lv,0);
      AppendInt2Listv(lv,siginfo.size);
    }
  } 
}








